#include <stdlib.h>
#include <string.h>
#include <errlog.h>
#include <epicsAssert.h>
#include <epicsThread.h>
#include <iocsh.h>
#include <registryFunction.h>
#include "imca-epics-unit.h"
#include "imca-ca-utils.h"
#include "dliEpcrLib.h"
#include "dliEpcrTestLib.h"
#include <epicsExport.h>

static EpicsUnitContext euc;
static char deviceAsynPortName[128];

static void testParseOutletStates01() {
  /* Basic. Empty response from PC8000. */
  short states[8];
  int status;

  status = parseOutletStates(DLI_MODEL_PC8000, "", states);
  failIf(&euc, status == 0);
}
static void testParseOutletStates02() {
  /* Basic. Normal response from PC8000. */
  short states[8];
  int status;
  char response[8192];

  response[0] = '\0';
  strcat(response,
    "<!-- heading rows -->\n"
    "<tr bgcolor=\"#DDDDDD\">\n"
    "<th>#</th>\n"
    "<th align=left>Name</th>\n"
    "<th align=left>State</th>\n"
    "<th align=left>Last switched</th>\n"
    "<th align=left colspan=2>Action</th>\n"
    "</tr>\n"
    "<!-- /heading rows -->\n"
    "\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>1</td><td>PC1</td><td><b><font "
    "color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;</td><td><a "
    "href=outleton?1>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>2</td><td>PC2</td><td><b><font "
    "color=green>ON</font></b></td><td>09/18/2006 11:57:28&nbsp;</td><td><a "
    "href=outletoff?2>Switch OFF</a></td><td><a href=outletgl?2>Cycle</a>"
    "</td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>3</td><td>Router</td><td><b>"
    "<font color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;</td>"
    "<td><a href=outleton?3>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>4</td><td>AC</td><td><b><font "
    "color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;</td><td><a "
    "href=outleton?4>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>5</td><td>PC2</td><td><b><font "
    "color=green>ON</font></b></td><td>09/18/2006 11:57:28&nbsp;</td><td><a "
    "href=outletoff?5>Switch OFF</a></td><td><a href=outletgl?5>Cycle</a>"
    "</td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>6</td><td>Outlet 6 - Not Used"
    "</td><td><b><font color=red>OFF</font></b></td><td>09/18/2006 11:55:58"
    "&nbsp;</td><td><a href=outleton?6>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>7</td><td>Outlet 7 - Not Used"
    "</td><td><b><font color=red>OFF</font></b></td><td>09/18/2006 11:55:58"
    "&nbsp;</td><td><a href=outleton?7>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>8</td><td>LightsOutside</td>"
    "<td><b><font color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;"
    "</td><td><a href=outleton?8>Switch ON</a></td><td></td></tr>\n");

  status = parseOutletStates(DLI_MODEL_PC8000, response, states);
  failIf(&euc, status != 0);
  if (status == 0) {
    failIfIntNotEqual(&euc, 0, states[0]);
    failIfIntNotEqual(&euc, 1, states[1]);
    failIfIntNotEqual(&euc, 0, states[2]);
    failIfIntNotEqual(&euc, 0, states[3]);
    failIfIntNotEqual(&euc, 1, states[4]);
    failIfIntNotEqual(&euc, 0, states[5]);
    failIfIntNotEqual(&euc, 0, states[6]);
    failIfIntNotEqual(&euc, 0, states[7]);
  }
}
static void testParseOutletStates03() {
  /* Basic. Partial response from PC8000. */
  short states[8];
  int status;
  char response[8192];

  response[0] = '\0';
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>1</td><td>PC1</td><td><b><font "
    "color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;</td><td><a "
    "href=outleton?1>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>2</td><td>PC2</td><td><b><font "
    "color=green>ON</font></b></td><td>09/18/2006 11:57:28&nbsp;</td><td><a "
    "href=outletoff?2>Swit");

  status = parseOutletStates(DLI_MODEL_PC8000, response, states);
  failIf(&euc, status == 0);
}
static void testParseOutletStates04() {
  /* Basic. NULL response. */
  short states[8];
  int status;

  status = parseOutletStates(DLI_MODEL_PC8000, NULL, states);
  failIf(&euc, status == 0);
}
static void testParseOutletStates05() {
  /* Basic. Partial match. */
  short states[8];
  int status;
  char response[8192];

  response[0] = '\0';
  strcat(response, "<a href=outlet");

  status = parseOutletStates(DLI_MODEL_PC8000, response, states);
  failIf(&euc, status == 0);
}
static void testParseOutletStates06() {
  /* Basic. Partial response. */
  short states[8];
  int status;
  char response[8192];

  response[0] = '\0';
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>1</td><td>PC1</td><td><b><font "
    "color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;</td><td><a "
    "href=outleton?1>Switch ON</a></td><td></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>2</td><td>PC2</td><td><b><font "
    "color=red>OFF</font></b></td><td>09/18/2006 11:55:58&nbsp;</td><td><a "
    "href=outleton?2>Switch ON</a></td><td></td></tr>\n");

  status = parseOutletStates(DLI_MODEL_PC8000, response, states);
  failIf(&euc, status == 0);
}
static void testParseOutletStates07() {
  /* Basic. Empty response from EPCR2. */
  short states[8];
  int status;

  status = parseOutletStates(DLI_MODEL_EPCR2, "", states);
  failIf(&euc, status == 0);
}
static void testParseOutletStates08() {
  /* Basic. Normal response from EPCR2. */
  short states[8];
  int status;
  char response[8192];

  response[0] = '\0';
  strcat(response,
    "<!-- heading rows -->\n"
    "<tr bgcolor=\"#DDDDDD\">\n"
    "<th>#</th>\n"
    "<th align=left>Name</th>\n"
    "<th align=left>State</th>\n"
    "<th align=left colspan=2>Action</th>\n"
    "</tr>\n"
    "<!-- /heading rows -->\n"
    "\n");

  strcat(response,
    "<tr><th colspan=\"2\" align=\"center\" bgcolor=\"#ddccdd\">Bus A:</th>"
    "<th colspan=\"3\" align=\"left\" bgcolor=\"#ddccdd\">121V 0.1A<!-- RAW  "
    "VA=394 CA=1 VAH=121.119 CAH=0.07 --></th></tr><tr bgcolor=\"#F4F4F4\">"
    "<td align=center>1</td><td>DNS-1 Prim</td><td><b><font color=red>OFF"
    "</font></b></td><td><a  href=outlet?1=ON>Switch ON</a></td><td><!-- <a  "
    "href=outlet?1=CCL>Cycle</a> --></td></tr>\n");

  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>2</td> <td>DNS-2 Prim</td><td> "
    "<b><font color=green>ON</font></b></td><td> <a onclick=\"return confirm("
    "'You are going to switch outlet 2 [DNS-2 Prim] OFF. Continue?')\" "
    "href=outlet?2=OFF>Switch OFF</a> </td><td> <a onclick=\"return confirm("
    "'You are going to cycle outlet 2 [DNS-2 Prim]. Continue?')\" "
    "href=outlet?2=CCL>Cycle</a> </td></tr>\n");

  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>3</td> <td>LCR-1 Prim</td><td> "
    "<b><font color=green>ON</font></b></td><td> <font color=\"#CCCCCC\">"
    "LOCKED</font><!-- <a onclick=\"return confirm('You are going to switch "
    "outlet 3 [LCR-1 Prim] OFF. Continue?')\" href=outlet?3=OFF>Switch OFF"
    "</a> --> </td><td> <font color=\"#CCCCCC\">LOCKED</font><!-- <a onclick="
    "\"return confirm('You are going to cycle outlet 3 [LCR-1 Prim]. "
    "Continue?')\" href=outlet?3=CCL>Cycle</a> --> </td></tr>\n");

  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>4</td> <td>LCR-2 Prim</td><td> "
    "<b><font color=green>ON</font></b></td><td> <a onclick=\"return confirm("
    "'You are going to switch outlet 4 [LCR-2 Prim] OFF. Continue?')\" href="
    "outlet?4=OFF>Switch OFF</a> </td><td> <a onclick=\"return confirm('You "
    "are going to cycle outlet 4 [LCR-2 Prim]. Continue?')\" href=outlet?4="
    "CCL>Cycle</a> </td></tr>\n");

  strcat(response,
    "<tr><th colspan=\"2\" align=\"center\" bgcolor=\"#ddccdd\">Bus B:</th>"
    "<th colspan=\"3\" align=\"left\" bgcolor=\"#ddccdd\">121V 0.1A<!-- RAW  "
    "VB=392 CB=1 VBH=121.064 CBH=0.07 --></th></tr> <tr bgcolor=\"#F4F4F4\">"
    "<td align=center>5</td> <td>DNS-1 Bak</td><td> <b><font color=green>ON"
    "</font></b></td><td> <a onclick=\"return confirm('You are going to "
    "switch outlet 5 [DNS-1 Bak] OFF. Continue?')\" href=outlet?5=OFF>Switch "
    "OFF</a> </td><td> <a onclick=\"return confirm('You are going to cycle "
    "outlet 5 [DNS-1 Bak]. Continue?')\" href=outlet?5=CCL>Cycle</a> </td>"
    "</tr>\n");

  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>6</td> <td>DNS-2 Bak</td><td> "
    "<b><font color=green>ON</font></b></td><td> <a  href=outlet?6=OFF>"
    "Switch OFF</a> </td><td> <a  href=outlet?6=CCL>Cycle</a> </td></tr>\n");

  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>7</td> <td>LCR-1 Bak</td><td> "
    "<b><font color=red>OFF</font></b></td><td> <font color=\"#CCCCCC\">"
    "LOCKED</font><!-- <a  href=outlet?7=ON>Switch ON</a> --> </td><td> <!-- "
    "<a  href=outlet?7=CCL>Cycle</a> --> </td></tr>\n");

  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>8</td> <td>LCR-2 Bak</td><td> "
    "<b><font color=green>ON</font></b></td><td> <a onclick=\"return confirm("
    "'You are going to switch outlet 8 [LCR-2 Bak] OFF. Continue?')\" href="
    "outlet?8=OFF>Switch OFF</a> </td><td> <a onclick=\"return confirm('You "
    "are going to cycle outlet 8 [LCR-2 Bak]. Continue?')\" href=outlet?8=CCL"
    ">Cycle</a> </td></tr>\n");

  status = parseOutletStates(DLI_MODEL_EPCR2, response, states);
  failIf(&euc, status != 0);
  if (status == 0) {
    failIfIntNotEqual(&euc, 0, states[0]);
    failIfIntNotEqual(&euc, 1, states[1]);
    failIfIntNotEqual(&euc, 1, states[2]);
    failIfIntNotEqual(&euc, 1, states[3]);
    failIfIntNotEqual(&euc, 1, states[4]);
    failIfIntNotEqual(&euc, 1, states[5]);
    failIfIntNotEqual(&euc, 0, states[6]);
    failIfIntNotEqual(&euc, 1, states[7]);
  }
}
static void testParseOutletStates09() {
  /* Basic. Partial response from EPCR2. */
  short states[8];
  int status;
  char response[8192];

  response[0] = '\0';
  strcat(response,
    "<tr><th colspan=\"2\" align=\"center\" bgcolor=\"#ddccdd\">Bus A:</th>"
    "<th colspan=\"3\" align=\"left\" bgcolor=\"#ddccdd\">121V 0.1A<!-- RAW  "
    "VA=394 CA=1 VAH=121.119 CAH=0.07 --></th></tr><tr bgcolor=\"#F4F4F4\">"
    "<td align=center>1</td><td>DNS-1 Prim</td><td><b><font color=red>OFF"
    "</font></b></td><td><a  href=outlet?1=ON>Switch ON</a></td><td><!-- <a  "
    "href=outlet?1=CCL>Cycle</a> --></td></tr>\n");
  strcat(response,
    "<tr bgcolor=\"#F4F4F4\"><td align=center>2</td> <td>DNS-2 Prim</td><td> "
    "<b><font color=green>ON</font></b></td><td> <a onclick=\"return confirm("
    "'You are going to switch outlet 2 [DNS-2 Prim] OFF. Continue?')\" "
    "href=outlet?2=OFF>Swi");

  status = parseOutletStates(DLI_MODEL_EPCR2, response, states);
  failIf(&euc, status == 0);
}

static void testOutletControl01() {
  /* Basic. */
  dbr_enum_t state;

  imcaSetEpicsPvToChar(TURN_ON_OUTLET_1_PV, '1');
  epicsThreadSleep(0.5);
  imcaGetEnumFromEpicsPv(OUTLET_1_STATE_PV, &state);
  failIfIntNotEqual(&euc, OUTLET_ON, state);

  imcaSetEpicsPvToChar(TURN_OFF_OUTLET_1_PV, '1');
  epicsThreadSleep(0.5);
  imcaGetEnumFromEpicsPv(OUTLET_1_STATE_PV, &state);
  failIfIntNotEqual(&euc, OUTLET_OFF, state);
}

static void testOutletControl02() {
  /* Basic. 10 times in a row. */
  int i;
  dbr_enum_t state;

  for (i = 0; i < 10; i++) {
    imcaSetEpicsPvToChar(TURN_ON_OUTLET_1_PV, '1');
    epicsThreadSleep(0.5);
    imcaGetEnumFromEpicsPv(OUTLET_1_STATE_PV, &state);
    failIfIntNotEqual(&euc, OUTLET_ON, state);

    imcaSetEpicsPvToChar(TURN_OFF_OUTLET_1_PV, '1');
    epicsThreadSleep(0.5);
    imcaGetEnumFromEpicsPv(OUTLET_1_STATE_PV, &state);
    failIfIntNotEqual(&euc, OUTLET_OFF, state);
  }
}

static void runAllTests() {
  runTest(&euc, testParseOutletStates01);
  runTest(&euc, testParseOutletStates02);
  runTest(&euc, testParseOutletStates03);
  runTest(&euc, testParseOutletStates04);
  runTest(&euc, testParseOutletStates05);
  runTest(&euc, testParseOutletStates06);
  runTest(&euc, testParseOutletStates07);
  runTest(&euc, testParseOutletStates08);
  runTest(&euc, testParseOutletStates09);
  runTest(&euc, testOutletControl01);
  runTest(&euc, testOutletControl02);
}

static void initializeDeviceAsynPortName(const char *inDeviceAsynPortName) {
  assert(strlen(inDeviceAsynPortName) < sizeof(deviceAsynPortName));
  strncpy(deviceAsynPortName, inDeviceAsynPortName,
    sizeof(deviceAsynPortName) - 1);
  deviceAsynPortName[sizeof(deviceAsynPortName) - 1] = '\0';
}

void dliEpcrTestRunner(const char *inDeviceAsynPortName) {
  initializeDeviceAsynPortName(inDeviceAsynPortName);
  runTests(&euc, runAllTests);
}

static const iocshArg dliEpcrTestRunnerArg0 = { "deviceAsynPortName",
  iocshArgString };
static const iocshArg *const dliEpcrTestRunnerArgs[] = {
  &dliEpcrTestRunnerArg0 };
static const iocshFuncDef dliEpcrTestRunnerDef = {
  "dliEpcrTestRunner", 1, dliEpcrTestRunnerArgs };
static void dliEpcrTestRunnerCall(const iocshArgBuf *args) {
  dliEpcrTestRunner(args[0].sval);
}
static void dliEpcrTestRegistrar(void) {
  static int isFirstTime = 1;
  if (!isFirstTime) return;
  isFirstTime = 0;
  iocshRegister(&dliEpcrTestRunnerDef, dliEpcrTestRunnerCall);
}
epicsExportRegistrar(dliEpcrTestRegistrar);
